/*
 * ----------------------------------------------------------------
 * Night Light IRC Proxy - Connection Logging Functions
 * ----------------------------------------------------------------
 * Copyright (C) 1997-2007 Jonas Kvinge <jonas@night-light.net>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Last modified by:
 * Jonas Kvinge (25.11.2007)
 *
 */

#define CONN_LOG_C

#define NEED_SYS_TYPES_H 1		/* Extra types */
#define NEED_SYS_PARAM_H 1		/* Some systems need this */
#define NEED_LIMITS_H 0			/* Kernel limits */
#define NEED_STDARG_H 1			/* va_list, etc */
#define NEED_ERRNO_H 1			/* errno */
#define NEED_CTYPE_H 1			/* isdigit(), etc */
#define NEED_NETINET_IN_H 1		/* in_addr, sockaddr_in, etc */
#define NEED_ARPA_INET_H 0		/* inet_ntoa(), inet_aton(), etc */
#define NEED_STDIO_H 1			/* Standard C UNIX functions */
#define NEED_STDLIB_H 1			/* malloc(), exit(), atoi(), etc */
#define NEED_TIME_H 1			/* time(), etc */
#define NEED_SYSCTL_H 0			/* sysctl(), etc */
#define NEED_SYS_STAT_H 1		/* chmod(), mkdir(), etc */
#define NEED_SYS_UIO_H 0		/* iovec, etc */
#define NEED_FCNTL_H 1			/* open(), creat(), fcntl(), etc */
#define NEED_SYS_IOCTL_H 0		/* ioctl(), etc */
#define NEED_SYS_FILIO_H 0		/* Solaris need this for ioctl(), etc */
#define NEED_UNISTD_H 1			/* Unix standard functions */
#define NEED_STRING_H 1			/* C string functions */
#define NEED_SIGNAL_H 0			/* Signal functions */
#define NEED_SYS_SOCKET_H 0		/* Socket functions */
#define NEED_NETDB_H 0			/* Network database functions */
#define NEED_ARPA_NAMESER_H 0		/* Nameserver definitions */
#define NEED_GETUSERPW_HEADERS 0 	/* Functions to retrive system passwords */
#define NEED_ARES 0			/* Functions needed for ares */
#define NEED_SSL 1			/* Needed for SSL support */

#include "includes.h"
#include "irc.h"

#include "conf.h"

#include "conn.h"
#include "conn_log.h"

extern struct Conf_Struct ConfS;
extern unsigned short int Root;

/* CONN_LOG FUNCTION - JONAS (18.07.2001) */

unsigned short int conn_log(struct Conn_Struct *ConnS, const char *const ChanPT, const char *const NickPT, const char *const UserPT, const char *const HostPT, const char *const NUHPT, const char *const CommandPT, const char *const ParamPT, const char *const MessagePT) {

  signed long int Result = 0;
  char *HomeDirPT = NULL;
  char LogFileName[FILELEN+1] = "";
  char LogPath[FILELEN+1] = "";
  char LogFileNamePath[FILELEN+1] = "";
  FILE *FilePT = NULL;
  struct tm *TMPT = NULL;
  char Date[LINELEN+1] = "";

  assert(ConnS != NULL);

  Conn_ClearLogHomeDir(ConnS);

  if ((ConfS.UnixPasswd == TRUE) && (Root == TRUE)) {
    HomeDirPT = sysgethomedirfromuser(ConnS->User);
    if (HomeDirPT == NULL) {
      time_t Duration = (NOW - ConnS->LogError);
      if (Duration > 3600) { sysprint(BITMASK_ERROR, "Unable to log messages for connection %s: Unable to retrive home directory for user \"%s\"", ConnS->Name, ConnS->User); }
      Conn_SetLogError(ConnS);
      ConnS->LogError = NOW;
      return(ERROR);
    }
    if (sysseteuidbyuser(ConnS->User) != SUCCESS) {
      time_t Duration = (NOW - ConnS->LogError);
      if (Duration > 3600) { sysprint(BITMASK_ERROR, "Unable to log messages for connection %s: Unable to set effective user ID to \"%s\"", ConnS->Name, ConnS->User); }
      Conn_SetLogError(ConnS);
      ConnS->LogError = NOW;
      return(ERROR);
    }
    if (access(HomeDirPT, X_OK) != SUCCESS) {
      time_t Duration = (NOW - ConnS->LogError);
      if (Duration > 3600) { sysprint(BITMASK_ERROR, "Unable to log messages for connection %s: Unable to access home directory for user \"%s\"", ConnS->Name, ConnS->User); }
      Conn_SetLogError(ConnS);
      ConnS->LogError = NOW;
      sysseteuidnormal();
      return(ERROR);
    }
    Conn_SetLogHomeDir(ConnS);
    snprintf(LogPath, FILELEN+1, "%s/ircproxy-logs", HomeDirPT);
  }
  else { snprintf(LogPath, FILELEN+1, "%s", ConfS.LogPath); }

  if (access(LogPath, X_OK) == ERROR) {
    Result = mkdir(LogPath, (S_IRWXU|S_IXGRP|S_IXOTH));
    if ((Result <= ERROR) && (errno != EEXIST)) {
      sysprint(BITMASK_ERROR, "Unable to create directory %s: [%d] %s", LogPath, errno, strerror(errno));
      if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
      return(ERROR);
    }
  }

  TMPT = localtime(&NOW);
  strftime(Date, LINELEN, "%Y%m%d", TMPT);

  if (CommandPT == NULL) {
    snprintf(LogFileName, FILELEN+1, "%s-%s-%s.log", ConnS->Name, ConnS->ServerHostName, Date);
    strfilefix(LogFileName);
  }
  else {
    if (ChanPT == NULL) {
      snprintf(LogFileName, FILELEN+1, "%s-%s.log", ConnS->Name, Date);
      strfilefix(LogFileName);
      ConnS->FileName = strrealloc(ConnS->FileName, LogFileName);
    }
    else {
      snprintf(LogFileName, FILELEN+1, "%s-%s-%s.log", ConnS->Name, ChanPT, Date);
      strfilefix(LogFileName);
    }
  }

  snprintf(LogFileNamePath, FILELEN+1, "%s/%s", LogPath, LogFileName);

  FilePT = fopen(LogFileNamePath, "a");
  if (FilePT == NULL) {
    sysprint(BITMASK_ERROR, "Unable to open %s for writing: [%d] %s", LogFileNamePath, errno, strerror(errno));
    if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
    return(ERROR);
  }
  Result = chmod(LogFileNamePath, S_IRUSR|S_IWUSR);
  if (Result <= ERROR) {
    sysprint(BITMASK_ERROR, "Unable to set file permission for %s: [%d] %s", LogFileNamePath, errno, strerror(errno));
    fclose(FilePT);
    if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
    return(ERROR);
  }

  if (CommandPT == NULL) { fprintf(FilePT, "%s *** %s%s", dtstamp(), MessagePT, LINEFEED); }
  else {
    assert(CommandPT != NULL);
    assert(NUHPT != NULL);
    assert(NickPT != NULL);
    assert(UserPT != NULL);
    assert(HostPT != NULL);
    if (strcasecmp(CommandPT, "PRIVMSG") == FALSE) { fprintf(FilePT, "%s <%s> %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
    else if (strcasecmp(CommandPT, "NOTICE") == FALSE) { fprintf(FilePT, "%s -%s- %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
    else if (strcasecmp(CommandPT, "ACTION") == FALSE) { fprintf(FilePT, "%s * %s %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
    else if (strcasecmp(CommandPT, "CTCP") == FALSE) {
      if (MessagePT == NULL)  { fprintf(FilePT, "%s %s CTCP %s%s", dtstamp(), NUHPT, ParamPT, LINEFEED); }
      else { fprintf(FilePT, "%s %s CTCP %s %s%s", dtstamp(), NUHPT, ParamPT, MessagePT, LINEFEED); }
    }
    else if (ChanPT != NULL) {
      if (strcasecmp(CommandPT, "NICK") == FALSE) { fprintf(FilePT, "%s %s (%s@%s) is now known as %s%s", dtstamp(), NickPT, UserPT, HostPT, ParamPT, LINEFEED); }
      else if (strcasecmp(CommandPT, "QUIT") == FALSE) {
        if (MessagePT == NULL) { fprintf(FilePT, "%s %s quits%s", dtstamp(), NUHPT, LINEFEED); }
        else { fprintf(FilePT, "%s %s quits: %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
      }
      else if (strcasecmp(CommandPT, "MODE") == FALSE) { fprintf(FilePT, "%s %s sets mode: %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
      else if (strcasecmp(CommandPT, "JOIN") == FALSE) { fprintf(FilePT, "%s %s has joined %s%s", dtstamp(), NUHPT, ChanPT, LINEFEED); }
      else if (strcasecmp(CommandPT, "PART") == FALSE) {
        if (MessagePT == NULL) { fprintf(FilePT, "%s %s has left %s%s", dtstamp(), NUHPT, ChanPT, LINEFEED); }
        else { fprintf(FilePT, "%s %s has left %s: %s%s", dtstamp(), NUHPT, ChanPT, MessagePT, LINEFEED); }
      }
      else if (strcasecmp(CommandPT, "KICK") == FALSE) { fprintf(FilePT, "%s %s was kicked by %s: %s%s", dtstamp(), ParamPT, NUHPT, MessagePT, LINEFEED); }
      else if (strcasecmp(CommandPT, "TOPIC") == FALSE) { fprintf(FilePT, "%s %s changed topic to: '%s'%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
    }
  }

  fclose(FilePT);

  if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }

  return(SUCCESS);

}

/* CONN_RAWLOG FUNCTION - JONAS (01.07.2000) */

void conn_rawlog(struct Conn_Struct *ConnS, const char *const LinePT, ...) {

  char Line[IRCMSGLEN+1] = "";
  va_list Args = { 0 };

  assert(ConnS != NULL);
  assert(LinePT != NULL);

  va_start(Args, LinePT);
  vsnprintf(Line, IRCMSGLEN+1, LinePT, Args);
  va_end(Args);

  conn_log(ConnS, NULL, NULL, NULL, NULL, NULL, NULL, NULL, Line);

}
